#    PMD to Collada converter, PMD_to_Collada.py
#
#    Converts 3d graphic files in the PMD format to the COLLADA XML-based
#    format
#
#    (C) 2010 by Matthew Minton 8/13/2010, v2.1
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Usage (hard way): Edit this file with a new PMD file path and run or,

# Usage (easy way): Import pmd2collada into your own script that lists 
# the PMD files you want to convert. Feed a new instance of PMDtoCollada the full
# path and watch it go. PMDtoCollada will convert the PMD files to DAE and 
# put them in the same directory as the source file.

import struct, sys, os, errno, datetime, easyxml
import xml.dom.minidom as xml
from vector_defs import *
from pmd2collada_defs import *

# global constants
TAB = '\t'
NL = '\n'

class PMD:
    """
        The main class containing all the elements of the PMD format
    """
    def __init__(self):
        self.magic = ''
        self.version = None
        self.datasize = None
        self.dataleft = None
        self.numVertices = 0
        self.vertices = []
        self.numFaces = 0
        self.faces = []
        self.numBones = 0
        self.bones = []
        self.numPropPoints = 0
        self.propPoints = []

    def write(self):
        text = self.magic + '\n' + \
               str(self.version) + '\n' + \
               str(self.datasize) + '\n'
        for vert in self.vertices:
            text += ('{0}: '.format(self.vertices.index(vert)) + vert.write())
        for face in self.faces:
            text += ('{0}: '.format(self.faces.index(face)) + face.write())
        for bone in self.bones:
            text += ('{0}: '.format(self.bones.index(bone)) + bone.write())
        for pp in self.propPoints:
            text += ('{0}: '.format(self.propPoints.index(pp)) + pp.write())
        return text
    
    def writeXML(self):
        pass

class Vertex:
    """
        Represents a single vertex in the PMD format.
    """
    def __init__(self):
        self.position = Vector3D() #  inst. of Vector3D
        self.normal = Vector3D() # inst of Vector3D
        self.uv = Vector2D() # inst of Vector2D
        self.blends = VertexBlend() # inst. of VertexBlend

    def write(self):
        text = self.position.write() + '\n' + \
               self.normal.write() + '\n' + \
               '{0} {1}'.format(self.u,self.v) + '\n' + \
               self.blends.write() + '\n'
        return text               

class VertexBlend:
    """
        Contains bone weighting for a given vertex
    """
    def __init__(self,bones=[],weights=[]):
        self.bones = bones # 4B
        self.weights = weights #4f

    def write(self):
        return '{0} {1}'.format(str(self.bones),str(self.weights))

class Face:
    """
        Stores the vertex ids that define a face
    """
    def __init__(self,verts=[0,0,0]):
        self.vertices = verts # 3B

    def write(self):
        return '{0} {1} {2}'.format(self.vertices[1],self.vertices[0],self.vertices[2])

class Bone:
    """
        Represents a single bone of a given skeleton
    """
    def __init__(self,trans=[0.0,0.0,0.0],rot=[0.0,0.0,0.0,0.0]):
        self.translation = Vector3D(trans) # inst of Vector3D
        self.rotation = Quaterion(rot) # inst of Quaterion

    def write(self):
        text = self.translation.write() + '\n' + \
               self.rotation.write() + '\n'
        return text

class PropPoint:
    """
        Represents an attachment point to a rigged mesh
    """
    def __init__(self):
        self.nameLength = 0
        self.name = ''
        self.translation = Vector3D()
        self.rotation = Quaterion()
        self.bone = 0 # 1B

    def write(self):
        text = self.name + '\n' + \
               self.translation.write() + '\n' + \
               self.rotation.write() + '\n' + \
               str(self.bone) + '\n'
        return text


class PMDtoCollada:
    """
        Converts a PMD mesh resource file to the XML-based Collada format

        This does all the heavy lifting, and should be the class used by external scripts.
        Feed it a filename (be nice, give it food it won't hork back up at you) and watch it go.
    """
    def __init__(self,filename):
        self.pmd = PMD()
        self.filename = os.path.split(filename)[1]
        self.pathname = os.path.split(filename)[0]
        self.filerootname = self.filename.split('.')[0]
        self.main()
       
    def extractPMDData(self,datafile):
        """
            Extracts binary PMD data and builds an instance of class PMD
            
            @datafile: the binary data file, open for reading
        """
        metadata = struct.unpack(header_format,
                                 datafile.read(header_length))
        self.pmd.magic = metadata[0]
        self.pmd.version = metadata[1]
        self.pmd.datasize = metadata[2]
        # Vertices
        self.pmd.numVertices = struct.unpack(vertex_header,
                                        datafile.read(vheader_length))[0]
        for i in range(self.pmd.numVertices):
            vert = Vertex()
            raw = struct.unpack(vertex_format,datafile.read(vertex_length))
            vert.position = Vector3D([round(x,6) for x in raw[0:3]])
            vert.normal = Vector3D([round(x,6) for x in raw[3:6]])
            vert.uv = Vector2D([round(x,6) for x in raw[6:8]])
            vert.blends = VertexBlend(raw[8:12],raw[12:16])
            self.pmd.vertices.append(vert)

        # Faces
        self.pmd.numFaces = struct.unpack(face_header,datafile.read(fheader_length))[0]
        for i in range(self.pmd.numFaces):
            raw = struct.unpack(face_format,datafile.read(face_length))
            face = Face(raw[0:3])
            self.pmd.faces.append(face)

        # Bones
        self.pmd.numBones = struct.unpack(bone_header,datafile.read(bheader_length))[0]
        for i in range(self.pmd.numBones):
            raw = struct.unpack(bone_format,datafile.read(bone_length))
            bone = Bone(raw[0:3],raw[3:7])
            self.pmd.bones.append(bone)


    def output(self):
        fout = open(r'd:\pmd_out_v2.txt','w')
        fout.write(self.pmd.write())
        fout.write('\n')
        fout.close()
        
    def outputCollada(self):
        try:
            savefolder = ('\\Converted')    
            os.makedirs(self.pathname + savefolder)
        except OSError, e:
            if e.errno != errno.EEXIST:
                raise
        fileName = "{0}\\{1}.dae".format(self.pathname + savefolder,self.filerootname)
        try:
            daeout = open(fileName,'w')
        except Exception as error:
            print error.msg
            print 'Export file failed to open, exiting...'
            sys.exit(0)
        now = datetime.datetime.now()
        doc = xml.Document()
        collada = easyxml.createElement(doc,'COLLADA',{'version':'1.4.0','xmlns':'http://www.collada.org/2005/11/COLLADASchema'})
        doc.appendChild(collada)
        asset = easyxml.createElement(doc,'asset',parent=collada)
        contributor = easyxml.createElement(doc,'contributor',parent=asset)
        author = easyxml.createElement(doc,'author',{},
            'PMD to Collada Coverter by Matthew Minton (c) 2010',
            parent=contributor)
        authoringTool = easyxml.createElement(doc,'authoring_tool',{},
            'pmd2collada.py, v1.0',
            parent=contributor)
        comments = easyxml.createElement(doc,'comments',parent=contributor)
        cp = easyxml.createElement(doc,'copyright',parent=contributor)
        source_data = easyxml.createElement(doc,'course_data',parent=contributor)
        created = easyxml.createElement(doc,'created',text=now.strftime('%Y-%m-%dT%H:%M:%S'),
            parent=asset)
        modified = easyxml.createElement(doc,'modified',text=now.strftime('%Y-%m-%dT%H:%M:%S'),
            parent=asset)
        unit = easyxml.createElement(doc,'unit',{'meter':'0.01','name':'centimeter'},parent=asset)
        upAxis = easyxml.createElement(doc,'up_axis',text='Y_UP',parent=asset)
        libGeo = easyxml.createElement(doc,'library_geometries',parent=collada)
        geo = easyxml.createElement(doc,'geometry',{'id':'{0}-Geometry'.format(self.filerootname),
            'name':'{0}-Geometry'.format(self.filerootname)},parent=libGeo)
        mesh = easyxml.createElement(doc,'mesh',parent=geo)
        posSource = easyxml.createElement(doc,'source',{'id':'{0}-Geometry-Position'.format(self.filerootname)},parent=mesh)
        posFloatArray = easyxml.createElement(doc,'float_array',{'count':'{0}'.format(self.pmd.numVertices*3),
            'id':'{0}-Geometry-Position-array'.format(self.filerootname)},
            '\n\t\t\t\t\t\t'.join([vert.position.write() for vert in self.pmd.vertices]),parent=posSource)
        posTechCommon = easyxml.createElement(doc,'technique_common',parent=posSource)
        posTechAccessor = easyxml.createElement(doc,'accessor',{'count':'{0}'.format(self.pmd.numVertices),
            'source':'{0}-Geometry-Position-array'.format(self.filerootname),'stride':'3'},parent=posTechCommon)
        posParamX = easyxml.createElement(doc,'param',{'type':'float','name':'X'},parent=posTechAccessor)
        posParamY = easyxml.createElement(doc,'param',{'type':'float','name':'Y'},parent=posTechAccessor)
        posParamZ = easyxml.createElement(doc,'param',{'type':'float','name':'Z'},parent=posTechAccessor)
        normSource = easyxml.createElement(doc,'source',{'id':'{0}-Geometry-Normals'.format(self.filerootname)},parent=mesh)
        normFloatArray = easyxml.createElement(doc,'float_array',{'count':'{0}'.format(self.pmd.numVertices*3),
            'id':'{0}-Geometry-Normal-array'.format(self.filerootname)},
            '\n\t\t\t\t\t\t'.join([vert.normal.write() for vert in self.pmd.vertices]),parent=normSource)
        normTechCommon = easyxml.createElement(doc,'technique_common',parent=normSource)
        normTechAccessor = easyxml.createElement(doc,'accessor',{'count':'{0}'.format(self.pmd.numVertices),
            'source':'{0}-Geometry-Normal-array'.format(self.filerootname),'stride':'3'},parent=normTechCommon)
        normParamX = easyxml.createElement(doc,'param',{'type':'float','name':'X'},parent=normTechAccessor)
        normParamY = easyxml.createElement(doc,'param',{'type':'float','name':'Y'},parent=normTechAccessor)
        normParamZ = easyxml.createElement(doc,'param',{'type':'float','name':'Z'},parent=normTechAccessor)
        uvSource = easyxml.createElement(doc,'source',{'id':'{0}-Geometry-UV'.format(self.filerootname)},parent=mesh)
        uvFloatArray = easyxml.createElement(doc,'float_array',{'count':'{0}'.format(self.pmd.numVertices*2),
            'id':'{0}-Geometry-UV-array'.format(self.filerootname)},
            '\n\t\t\t\t\t\t'.join([vert.uv.write() for vert in self.pmd.vertices]),parent=uvSource)
        uvTechCommon = easyxml.createElement(doc,'technique_common',parent=uvSource)
        uvTechAccessor = easyxml.createElement(doc,'accessor',{'count':'{0}'.format(self.pmd.numVertices),
            'source':'{0}-Geometry-UV-array'.format(self.filerootname),'stride':'2'},parent=uvTechCommon)
        uvParamX = easyxml.createElement(doc,'param',{'type':'float','name':'U'},parent=uvTechAccessor)
        uvParamY = easyxml.createElement(doc,'param',{'type':'float','name':'V'},parent=uvTechAccessor)
        vertices = easyxml.createElement(doc,'vertices',{'id':'{0}-Geometry-Vertex'.format(self.filerootname)},parent=mesh)
        vertInput = easyxml.createElement(doc,'input',{'semantic':'POSITION',
            'source':'#{0}-Geometry-Position'.format(self.filerootname)},parent=vertices)
        triangles = easyxml.createElement(doc,'triangles',
            {'material':'','count':'{0}'.format(self.pmd.numFaces)},parent=mesh)
        triInput1 = easyxml.createElement(doc,'input',{'offset':'0',
            'semantic':'VERTEX','source':'#{0}-Geometry-Vertex'.format(self.filerootname)},parent=triangles)
        triInput2 = easyxml.createElement(doc,'input',{'offset':'0',
            'semantic':'NORMAL','source':'#{0}-Geometry-Normals'.format(self.filerootname)},parent=triangles)
        triInput3 = easyxml.createElement(doc,'input',{'offset':'0',
            'semantic':'TEXCOORD','source':'#{0}-Geometry-UV'.format(self.filerootname)},parent=triangles)
        faceArray = easyxml.createElement(doc,'p',text='\n\t\t\t\t\t\t'.join([face.write() for face in self.pmd.faces]),parent=triangles)
        libVisScenes = easyxml.createElement(doc,'library_visual_scenes',parent=collada)
        visualScene = easyxml.createElement(doc,'visual_scene',{'id':'Scene','name':'Scene'})
        libVisScenes.appendChild(visualScene)
        node = easyxml.createElement(doc,'node',{'layer':'L1','id':'{0}'.format(self.filerootname),
            'name':'{0}'.format(self.filerootname)},parent=visualScene)
        rotate = easyxml.createElement(doc,'rotate',{'sid':'rotateX'},text='1 0 0 90.000000',parent=node)
        instGeo = easyxml.createElement(doc,'instance_geometry',
            {'url':'#{0}-Geometry'.format(self.filerootname)},parent=node)
        scene = easyxml.createElement(doc,'scene',parent=collada)
        instVisScene = easyxml.createElement(doc,'instance_visual_scene',{'url':'#Scene'},parent=scene)
        daeout.write(doc.toprettyxml())
        daeout.close()

    def main(self):
        try:
            datafile = open('{0}\\{1}'.format(self.pathname,self.filename),'rb')
        except Exception as error:
            print error
            print 'Input file failed to open, exiting...'
            sys.exit(0)
        else:
            print 'Starting data extraction on {0}...'.format(self.filerootname)
            self.extractPMDData(datafile)
            datafile.close()
            print 'Data extracted, writing {0}.dae...'.format(self.filerootname)
            self.outputCollada()
            print 'Complete!'

if __name__ == '__main__':
    testfile = r'D:\Profiles\b28670\My Documents\Personal\Python\Projects\PMD to Collada\celt_helmet_m.pmd'
    convert = PMDtoCollada(testfile)

